#include "ShmemQueue.h"
#include <stdio.h>

// Function name	: ShmemQueue::ShmemQueue
// Description	    : 
// Return type		: 
ShmemQueue::ShmemQueue()
{
	m_dwSize = 0;
	m_pBase = NULL;
	m_hMapping = NULL;
	m_hQMutex = NULL;
	m_hMsgAvailableEvent = NULL;
	m_hQEmptyEvent = NULL;
}

// Function name	: ShmemQueue::~ShmemQueue
// Description	    : 
// Return type		: 
ShmemQueue::~ShmemQueue()
{
	if (m_hMapping != NULL)
	{
		if (m_pBase != NULL)
			UnmapViewOfFile(m_pBase);
		CloseHandle(m_hMapping);
	}
	if (m_hQMutex != NULL)
		CloseHandle(m_hQMutex);
	if (m_hMsgAvailableEvent != NULL)
		CloseHandle(m_hMsgAvailableEvent);
	if (m_hQEmptyEvent != NULL)
		CloseHandle(m_hQEmptyEvent);
	m_dwSize = 0;
	m_pBase = NULL;
	m_hMapping = NULL;
	m_hQMutex = NULL;
	m_hMsgAvailableEvent = NULL;
	m_hQEmptyEvent = NULL;
}

// Function name	: ShmemQueue::Init
// Description	    : 
// Return type		: bool 
// Argument         : char *name
// Argument         : unsigned long size
bool ShmemQueue::Init(char *name, unsigned long size)
{
	bool bFirst = true;

	// Create a mapping from the page file
	m_hMapping = CreateFileMapping(
		INVALID_HANDLE_VALUE,
		NULL,
		PAGE_READWRITE,
		0, size,
		name);

	// If the mapping already exists then we are attaching to it, not creating it
	if (GetLastError() == ERROR_ALREADY_EXISTS)
		bFirst = false;

	if (m_hMapping == NULL)
		return false;

	// Map the file and save the pointer to the base of the mapped file
	m_pBase = MapViewOfFile(
		m_hMapping,
		FILE_MAP_WRITE,
		0,0,
		size);

	if (m_pBase == NULL)
	{
		CloseHandle(m_hMapping);
		m_hMapping = NULL;
		return false;
	}

	m_pEnd = (LPBYTE)m_pBase + size;
	m_dwSize = size;
	// max message = total mapped space - head and tail pointer - one message header
	m_dwMaxMsgSize = size - (2*sizeof(ShmemQueueHeader*)) - sizeof(ShmemQueueHeader);

	// If this process is creating the mapping, then set up the head and tail pointers
	if (bFirst)
	{
		((unsigned long*)m_pBase)[0] = 0;
		((unsigned long*)m_pBase)[1] = 2*sizeof(unsigned long);
	}

	// Create or Attach to the synchronization handles for this queue
	char pszMutexName[200], pszEventName[200], pszEvent2Name[200];

	sprintf(pszMutexName, "%s.mutex", name);
	sprintf(pszEventName, "%s.event", name);
	sprintf(pszEvent2Name, "%s.event2", name);

	m_hQMutex = CreateMutex(NULL, FALSE, pszMutexName);
	if (m_hQMutex == NULL)
	{
		CloseHandle(m_hMapping);
		m_hMapping = NULL;
		return false;
	}

	m_hMsgAvailableEvent = CreateEvent(NULL, TRUE, FALSE, pszEventName);
	if (m_hMsgAvailableEvent == NULL)
	{
		CloseHandle(m_hMapping);
		CloseHandle(m_hQMutex);
		m_hMapping = NULL;
		m_hQMutex = NULL;
		return false;
	}

	m_hQEmptyEvent = CreateEvent(NULL, TRUE, FALSE, pszEvent2Name);
	if (m_hQEmptyEvent == NULL)
	{
		CloseHandle(m_hMapping);
		CloseHandle(m_hQMutex);
		CloseHandle(m_hMsgAvailableEvent);
		m_hMapping = NULL;
		m_hQMutex = NULL;
		m_hMsgAvailableEvent = NULL;
		return false;
	}

	return true;
}

// Function name	: ShmemQueue::Insert
// Description	    : 
// Return type		: bool 
// Argument         : unsigned char *buffer
// Argument         : unsigned int length
// Argument         : int tag
// Argument         : int from
bool ShmemQueue::Insert(unsigned char *buffer, unsigned int length, int tag, int from)
{
	ShmemQueueHeader *pTail;

	if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
		return false;

	if (length > m_dwMaxMsgSize)
	{
		ReleaseMutex(m_hQMutex);
		return false;
	}

	// Wait for a contiguous block large enough to hold the data 
	while ((unsigned long)m_pEnd - (unsigned long)m_pBase - ((unsigned long *)m_pBase)[1] - (unsigned long)sizeof(ShmemQueueHeader) < length)
	{
		ReleaseMutex(m_hQMutex);
		if (WaitForSingleObject(m_hQEmptyEvent, INFINITE) != WAIT_OBJECT_0)
			return false;
		if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
			return false;
	}

	// Read the tail pointer
	pTail = (ShmemQueueHeader *)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
	// If the head pointer is 0, set it to the tail
	if ( *((unsigned long *)m_pBase) == 0 )
		*((unsigned long *)m_pBase) = (unsigned long)((LPBYTE)pTail - (LPBYTE)m_pBase);

	// Set the state and advance the tail offset
	pTail->state = SHMEM_Q_BEING_WRITTEN;
	// Advance the tail pointer
	((unsigned long *)m_pBase)[1] = (unsigned long)((unsigned long)pTail + length + sizeof(ShmemQueueHeader) - (unsigned long)m_pBase);
	ReleaseMutex(m_hQMutex);

	// Write the header
	pTail->tag = tag;
	pTail->from = from;
	pTail->length = length;
	pTail->next_offset = (sizeof(ShmemQueueHeader) + length);
	
	// Copy the data
	memcpy((LPBYTE)pTail + sizeof(ShmemQueueHeader), buffer, length);


	if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
		return false;
	// Signal data has arrived and release the mutex
	pTail->state = SHMEM_Q_AVAIL_FOR_READ;
	SetEvent(m_hMsgAvailableEvent);
	ResetEvent(m_hQEmptyEvent);
	ReleaseMutex(m_hQMutex);

	return true;
}

// Function name	: ShmemQueue::RemoveNext
// Description	    : 
// Return type		: bool 
// Argument         : unsigned char *buffer
// Argument         : unsigned int *length
// Argument         : int *tag
// Argument         : int *from
bool ShmemQueue::RemoveNext(unsigned char *buffer, unsigned int *length, int *tag, int *from)
{
	ShmemQueueHeader *pTail, *pHead, *pMessage;

	// Get the next available entry in the queue
	while(true)
	{
		// Get the queue mutex
		if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
		{
			printf("Wait for QMutex failed\n");
			return false;
		}
		
		// Wait for the queue to not be empty
		while (((unsigned long*)m_pBase)[0] == 0)
		{
			ReleaseMutex(m_hQMutex);
			if (WaitForSingleObject(m_hMsgAvailableEvent, INFINITE) != WAIT_OBJECT_0)
			{
				printf("Wait for MsgAvailableEvent failed\n");
				return false;
			}
			if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
			{
				printf("Wait for QMutex failed\n");
				return false;
			}
		}
		
		// Search the queue for the next available entry
		pMessage = (ShmemQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[0] );
		pTail = (ShmemQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
		
		while ((pMessage->state == SHMEM_Q_BEING_READ) && (pMessage < pTail))
			pMessage = (ShmemQueueHeader*)((LPBYTE)pMessage + pMessage->next_offset);
		
		// If we haven't reached the tail and the element we are on is not currently being written, then successfully break out of this loop
		if ( (pMessage < pTail) && (pMessage->state != SHMEM_Q_BEING_WRITTEN) )
			break;

		// All messages are being read or the next message in order is not ready. I need to reset MsgAvailableEvent, wait for it to be signalled and then start over
		ResetEvent(m_hMsgAvailableEvent);
		ReleaseMutex(m_hQMutex);
		WaitForSingleObject(m_hMsgAvailableEvent, INFINITE);
	}

	// Check that the buffer provided is large enough to hold the data
	if (pMessage->length > *length)
	{
		printf("length %d > *length %d\n", pMessage->length, *length);
		ReleaseMutex(m_hQMutex);
		return false;
	}

	// Mark the message as being read
	pMessage->state = SHMEM_Q_BEING_READ;
	ReleaseMutex(m_hQMutex);

	// Read the data from the message
	*tag = pMessage->tag;
	*from = pMessage->from;
	*length = pMessage->length;
	memcpy(buffer, (LPBYTE)pMessage + sizeof(ShmemQueueHeader), pMessage->length);

	if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
	{
		printf("Wait for QMutex failed\n");
		return false;
	}
	// Mark the message as having been read
	pMessage->state = SHMEM_Q_READ;

	// Update the head and tail pointers of the queue
	pHead = (ShmemQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[0] );
	pTail = (ShmemQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
		
	// Advance the head pointer over all the read messages
	while ((pHead->state == SHMEM_Q_READ) && (pHead < pTail))
		pHead = (ShmemQueueHeader*)((LPBYTE)pHead + pHead->next_offset);

	if (pHead >= pTail)
	{
		// When the head catches up to the tail, the queue is empty so reset the pointers and signal the queue empty
		((unsigned long*)m_pBase)[0] = 0;
		((unsigned long*)m_pBase)[1] = 2*sizeof(unsigned long);
		ResetEvent(m_hMsgAvailableEvent);
		SetEvent(m_hQEmptyEvent);
	}
	else
		((unsigned long*)m_pBase)[0] = (unsigned long)((LPBYTE)pHead - (LPBYTE)m_pBase);

	ReleaseMutex(m_hQMutex);

	return true;
}

// Function name	: ShmemQueue::RemoveNextInsert
// Description	    : 
// Return type		: bool 
// Argument         : MessageQueue *pMsgQueue
bool ShmemQueue::RemoveNextInsert(MessageQueue *pMsgQueue)
{
	ShmemQueueHeader *pTail, *pHead, *pMessage;

	// Get the next available entry in the queue
	while(true)
	{
		// Get the queue mutex
		if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
		{
			printf("Wait for QMutex failed\n");
			return false;
		}
		
		// Wait for the queue to not be empty
		while (((unsigned long*)m_pBase)[0] == 0)
		{
			ReleaseMutex(m_hQMutex);
			if (WaitForSingleObject(m_hMsgAvailableEvent, INFINITE) != WAIT_OBJECT_0)
			{
				printf("Wait for MsgAvailableEvent failed\n");
				return false;
			}
			if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
			{
				printf("Wait for QMutex failed\n");
				return false;
			}
		}
		
		// Search the queue for the next available entry
		pMessage = (ShmemQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[0] );
		pTail = (ShmemQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
		
		while ((pMessage->state == SHMEM_Q_BEING_READ) && (pMessage < pTail))
			pMessage = (ShmemQueueHeader*)((LPBYTE)pMessage + pMessage->next_offset);
		
		// If we haven't reached the tail and the element we are on is not currently being written, then successfully break out of this loop
		if ( (pMessage < pTail) && (pMessage->state != SHMEM_Q_BEING_WRITTEN) )
			break;

		// All messages are being read or the next message in order is not ready. I need to reset MsgAvailableEvent, wait for it to be signalled and then start over
		ResetEvent(m_hMsgAvailableEvent);
		ReleaseMutex(m_hQMutex);
		WaitForSingleObject(m_hMsgAvailableEvent, INFINITE);
	}

	MessageQueue::MsgQueueElement *pElement;
	if (pMessage->from == -1)
		return false;
	void *pBuffer = pMsgQueue->GetBuffer(pMessage->tag, pMessage->length, pMessage->from, &pElement);

	// Mark the message as being read
	pMessage->state = SHMEM_Q_BEING_READ;
	ReleaseMutex(m_hQMutex);

	// Read the data from the message
	memcpy(pBuffer, (LPBYTE)pMessage + sizeof(ShmemQueueHeader), pMessage->length);
	pMsgQueue->SetElementEvent(pElement);

	if (WaitForSingleObject(m_hQMutex, INFINITE) != WAIT_OBJECT_0)
	{
		printf("Wait for QMutex failed\n");
		return false;
	}
	// Mark the message as having been read
	pMessage->state = SHMEM_Q_READ;

	// Update the head and tail pointers of the queue
	pHead = (ShmemQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[0] );
	pTail = (ShmemQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
		
	// Advance the head pointer over all the read messages
	while ((pHead->state == SHMEM_Q_READ) && (pHead < pTail))
		pHead = (ShmemQueueHeader*)((LPBYTE)pHead + pHead->next_offset);

	if (pHead >= pTail)
	{
		// When the head catches up to the tail, the queue is empty so reset the pointers and signal the queue empty
		((unsigned long*)m_pBase)[0] = 0;
		((unsigned long*)m_pBase)[1] = 2*sizeof(unsigned long);
		ResetEvent(m_hMsgAvailableEvent);
		SetEvent(m_hQEmptyEvent);
	}
	else
		((unsigned long*)m_pBase)[0] = (unsigned long)((LPBYTE)pHead - (LPBYTE)m_pBase);

	ReleaseMutex(m_hQMutex);

	return true;
}
